from pyfakefs import fake_filesystem
from blackhc.project.utils import simple_storage
import os
import pytest


def test_save_and_load_pkl(fs: fake_filesystem.FakeFilesystem):
    test_obj = {"key": "value"}
    root = "/tmp/blackhc.project"
    fs.CreateDirectory(root)
    
    path = simple_storage.save_pkl(test_obj, "test", root=root)
    loaded_obj = simple_storage.load_pkl("test", root=root)
    
    assert test_obj == loaded_obj
    assert os.path.exists(f"{path}.meta.json")


def test_save_and_load_json(fs: fake_filesystem.FakeFilesystem):
    test_obj = {"key": "value"}
    root = "/tmp/blackhc.project"
    fs.CreateDirectory(root)
    
    path = simple_storage.save_json(test_obj, "test", root=root)
    loaded_obj = simple_storage.load_json("test", root=root)
    
    assert test_obj == loaded_obj
    assert os.path.exists(f"{path}.meta.json")


def test_save_and_load_pt(fs: fake_filesystem.FakeFilesystem):
    if simple_storage.torch is None:
        pytest.skip("torch is not available")
    
    test_obj = simple_storage.torch.tensor([1, 2, 3])
    root = "/tmp/blackhc.project"
    fs.CreateDirectory(root)
    
    path = simple_storage.save_pt(test_obj, "test", root=root)
    loaded_obj = simple_storage.load_pt("test", root=root)
    
    assert simple_storage.torch.equal(test_obj, loaded_obj)
    assert os.path.exists(f"{path}.meta.json")


def test_save_pkl_or_json(fs: fake_filesystem.FakeFilesystem):
    test_obj = {"key": "value"}
    root = "/tmp/blackhc.project"
    fs.CreateDirectory(root)
    
    path = simple_storage.save_pkl_or_json(test_obj, "test", root=root)
    loaded_obj = simple_storage.load("test", root=root)
    
    assert test_obj == loaded_obj
    assert os.path.exists(f"{path}.meta.json")


def test_collect_metadata(fs: fake_filesystem.FakeFilesystem):
    root = "/tmp/blackhc.project"
    fs.CreateDirectory(root)
    
    metadata = simple_storage.collect_metadata("test")
    assert "timestamp" in metadata
    assert "git" in metadata
    assert "wandb" in metadata
    assert "parts" in metadata


def test_cache_decorator(fs: fake_filesystem.FakeFilesystem):
    root = "/tmp/blackhc.project"
    fs.CreateDirectory(root)
    
    @simple_storage.cache(prefix_args=["arg1"], root=root, force_format="json")
    def test_function(arg1, arg2):
        return {"result": arg1 + arg2}
            
    # Test loading directly from cache
    with pytest.raises(FileNotFoundError):
        loaded_result = test_function.load(1, 2)
    
    result = test_function(1, 2)
    assert result == {"result": 3}
    
    cached_result = test_function(1, 2)
    assert cached_result == result
    
    # Test loading directly from cache
    loaded_result = test_function.load(1, 2)
    assert loaded_result == result
    
    # Test recomputing the result
    recomputed_result = test_function.recompute(1, 2)
    assert recomputed_result == result
    # We should have two subdirectories now
    assert len(os.listdir(test_function.get_prefix_path(1, 2))) == 2



"""
Logging execution paths to make it easy to create nested paths for logging and debugging.

This module provides a `LogPath` class (aliased as `xpath`) that allows for hierarchical
logging of execution paths. It's particularly useful for creating nested paths for
logging and debugging in complex applications.

Key features:
- Create nested execution steps with unique identifiers
- Track summary steps separately from regular steps
- Automatically measure and record execution time for each step
- Generate unique metric names based on the current execution path
- Maintain statistics for step execution times

Usage:
```
    from blackhc.project.utils.log_path import xpath

    with xpath.step("main_process"):
        # Do some work
        with xpath.step("sub_process"):
            # Do sub-process work
            metric_name = xpath.metric_name("accuracy")
            # Log metric_name with your logging system

    # Access statistics
    print(xpath.all_step_names)
    print(xpath.all_step_stats)
```

Classes:
    LogPathStep: Represents a single step in the execution path
    LogPathNodeStats: Stores statistics for a specific step
    LogPath: Main class for managing the execution path and logging

Note: This module uses ContextVar for thread-safe context management.
"""
import dataclasses
import warnings
from collections import defaultdict
from contextlib import contextmanager
from contextvars import ContextVar
from dataclasses import dataclass

from blackhc.project.utils.context_stopwatch import ContextStopwatch


@dataclass
class LogPathStep:
    step_name: str
    step_index: int
    unique_path: str
    is_summary_step: bool = False
    parent: "LogPathStep | None" = None
    used_names: dict[str, int] = dataclasses.field(
        default_factory=lambda: defaultdict(int)
    )

    def create_child(
        self, name, *, path_separator: str, is_summary_step: bool = False
    ) -> "LogPathStep":
        is_summary_step |= self.is_summary_step

        step_index = self.used_names[name]

        # Built a unique name that is fully indexed.
        if is_summary_step:
            unique_name = name

            if self.used_names[name] >= 1:
                unique_name = f"{name}+{step_index}"

                if step_index == 1:
                    # Log a warning that the name has been used before.
                    warnings.warn(
                        f"Sub name {name} has already been used in summary step. Use a unique name please!"
                    )
        else:
            unique_name = f"{name}[{step_index}]"

        # Increment the number of times the sub_name has been used.
        self.used_names[name] += 1

        if self.parent is None:
            step_name = name
            unique_path = unique_name
        else:
            step_name = f"{self.step_name}{path_separator}{name}"
            unique_path = f"{self.unique_path}{path_separator}{unique_name}"

        if is_summary_step:
            step_name = unique_path

        return LogPathStep(step_name, step_index, unique_path, is_summary_step, self)

    def metric_name(
        self, metric_name, *, path_separator: str, is_summary: bool = False
    ):
        if self.parent is None:
            return metric_name
        else:
            if is_summary:
                return f"{self.unique_path}{path_separator}{metric_name}"
            else:
                return f"{self.step_name}{path_separator}{metric_name}"


@dataclass
class LogPathNodeStats:
    count: int = 0
    first_time: float = float("nan")
    rest_time: float = float("nan")

    @property
    def mean_rest_time(self):
        if self.count == 0:
            return self.rest_time
        return self.rest_time / (self.count - 1)

    @property
    def mean_time(self):
        if self.count <= 1:
            return self.first_time
        return (self.first_time + self.rest_time) / self.count

    @property
    def total_time(self):
        if self.count <= 1:
            return self.first_time
        return self.first_time + self.rest_time

    def append_time(self, time):
        if self.count == 0:
            self.first_time = time
        else:
            if self.count == 1:
                self.rest_time = 0.0
            self.rest_time += time

        self.count += 1


class LogPath:
    """
    Log Execution Paths Manager.

    A simple class to make it easy to create nested paths for logging and debugging.
    """

    path_separator = "/"
    _ROOT: LogPathStep = LogPathStep("", 0, "")
    _current: ContextVar[LogPathStep] = ContextVar("current_xpath_node", default=_ROOT)
    all_summary_step_names = set()
    all_step_names = set()
    all_metrics = set()
    all_summary_metrics = set()
    all_step_stats = defaultdict(LogPathNodeStats)

    @classmethod
    @contextmanager
    def step(cls, name, is_summary_step: bool = False):
        """
        Create a new sub-path.

        Args:
            name: The name of the step.
            is_summary_step: If True, the step will be treated as a summary step.
                This means that a unique name will be generated for the step.
        """
        current = cls._current.get()
        child = current.create_child(
            name, path_separator=cls.path_separator, is_summary_step=is_summary_step
        )
        if child.is_summary_step:
            cls.all_summary_step_names |= {child.step_name}
        else:
            cls.all_step_names |= {child.step_name}
        token = cls._current.set(child)
        try:
            with ContextStopwatch() as stopwatch:
                yield child
            cls.all_step_stats[child.step_name].append_time(stopwatch.elapsed_time)
        finally:
            cls._current.reset(token)

    @classmethod
    @property
    def current_step_name(cls):
        """
        Get the current path.

        Returns:
            The current path.
        """
        return cls._current.get().step_name

    @classmethod
    @property
    def current_step_index(cls):
        """
        Get the current step index.

        Returns:
            The current step index.
        """
        return cls._current.get().step_index

    @classmethod
    @property
    def current_step_count(cls):
        """
        Get the current step count.

        Returns:
            The current step count.
        """
        return cls.all_step_stats[cls.current_step_name].count

    @classmethod
    @property
    def current_unique_path(cls):
        """
        Get the current unique path.

        Returns:
            The current unique path.
        """
        return cls._current.get().unique_path

    @classmethod
    def metric_name(cls, metric_name, is_summary: bool = False):
        """
        Get the full metric name.

        Args:
            metric_name:

        Returns:
            The full metric name.
        """
        metric_name = cls._current.get().metric_name(
            metric_name, path_separator=cls.path_separator, is_summary=is_summary
        )
        if is_summary:
            cls.all_summary_metrics |= {metric_name}
        else:
            cls.all_metrics |= {metric_name}
        return metric_name

    @classmethod
    def test_reset(cls):
        cls._current.set(cls._ROOT)
        cls._ROOT.used_names.clear()
        cls.all_step_names = set()
        cls.all_summary_step_names = set()
        cls.all_step_stats.clear()
        cls.all_metrics = set()
        cls.all_summary_metrics = set()


xpath = LogPath

# Limit exports
__all__ = ["xpath"]



"""
Instrumentation Module

This module provides utilities for instrumenting code to collect and manage runtime data.
It offers a flexible and thread-safe way to record, organize, and transform instrumentation
data across different scopes of execution.

Key components:
- Context: A dict-like class for storing and manipulating instrumentation data
- Instrumentation: A class that manages the collection and scoping of instrumentation data
- NullInstrumentation: A no-op version of Instrumentation for when instrumentation is disabled

Usage:
```
    from blackhc.project.utils.instrumentation import instrumentation

    with instrumentation.collect() as run_info:
        instrumentation.log(metric="value")
        with instrumentation.scope("subscope"):
            instrumentation.log(another_metric="another_value")
```

The module also supports disabling instrumentation via the NO_INSTRUMENTATION environment variable.
"""
import contextlib
import os
import threading
import types
import typing
from dataclasses import dataclass

try:
    import torch
except ImportError:
    raise ImportError("Package 'torch' is not installed. Please install it using 'pip install torch'")

T = typing.TypeVar("T")


class Context(dict[str, list]):
    """
    Wrapper class for scopes .
    """

    def walk(self) -> typing.Iterable[tuple[str, typing.Any]]:
        """Replay the collected instrumentation data.

        This returns an iterable of (name, value) pairs.
        """
        for key, values in self.items():
            for value in values:
                yield key, value

    def tree_walk(self) -> typing.Iterable[tuple[tuple[str], typing.Any]]:
        """Replay the collected instrumentation data.

        This returns an iterable of (path, value) pairs, where path is a tuple of keys.
        """
        for key, values in self.items():
            for value in values:
                yield (key,), value

                if isinstance(value, Context):
                    for sub_path, sub_value in value.tree_walk():
                        yield (key,) + sub_path, sub_value

    def transform_values(self, t_value: typing.Callable[[T], T]) -> "Context":
        """Transform the collected instrumentation data.

        This returns a new context with all values transformed by the given function.
        Recursively applies to sub-contexts.
        """
        return Context(
            {
                k: [
                    v.transform_values(t_value)
                    if isinstance(v, Context)
                    else t_value(v)
                    for v in vs
                ]
                for k, vs in self.items()
            }
        )

    def to_dict(self) -> dict:
        """Convert the collected instrumentation data to a dict.

        This returns a new context with all values transformed by the given function.
        Recursively applies to sub-contexts.
        """
        return {
            k: [v.to_dict() if isinstance(v, Context) else v for v in vs]
            for k, vs in self.items()
        }

    def to_numpy(self) -> "Context":
        """Convert the collected instrumentation data to a dict.

        This returns a new context with all values transformed by the given function.
        Recursively applies to sub-contexts.
        """
        return self.transform_values(
            t_value=lambda x: x.numpy(force=True) if isinstance(x, torch.Tensor) else x,
        )

    def to_simple_namespace(self) -> types.SimpleNamespace:
        """Convert the collected instrumentation data to a SimpleNamespace.

        This returns a new context with all values transformed by the given function.
        Recursively applies to sub-contexts.
        """
        return types.SimpleNamespace(
            **{
                k: [
                    v.to_simple_namespace() if isinstance(v, Context) else v for v in vs
                ]
                for k, vs in self.items()
            }
        )


@dataclass
class Instrumentation:
    """A simple helper that allows you to easily add instrumentation to your code.

    Example usage:

        ```
        from utils.instrumentation import instrument

        with instrument.collect() as run_info:
            # your code here
            instrument.record("loss", loss)

            with instrument("my_scope"):
                # your code here
                instrument.record("info", info)
        ```

        You can also use the `instrument` function as a decorator:

        ```
        @instrument("my_scope")
        def my_func():
            # your code here
            instrument.record("info", info)
        ```

        All scopes are stored as lists of SimpleNamespace objects
    """

    _thread_local: threading.local = threading.local()

    @property
    def _context(self) -> list[Context] | None:
        """Get the context instrumentation data"""
        return getattr(self._thread_local, "context", None)

    @property
    def is_active(self) -> bool:
        """Check if instrumentation is currently active."""
        return self._context is not None

    @contextlib.contextmanager
    def scope(self, key: str):
        """Create a new scope.

        If no data is being collected currently, this is a no-op.
        """
        outer = self._context
        if outer is None:
            yield
        else:
            inner = Context()
            self._thread_local.context = inner
            try:
                yield
                outer.setdefault(key, []).append(inner)
            finally:
                self._thread_local.context = outer

    def log(self, **kwargs):
        """Record a set of values.

        If no data is being collected, this is a no-op.
        """
        context = self._context
        if context is not None:
            for key, value in kwargs.items():
                context_value = value
                # detach if value is a torch.Tensor
                if isinstance(value, torch.Tensor):
                    context_value = value.detach()
                context.setdefault(key, []).append(context_value)

    def spy(self, **kwargs):
        """Record a single value.

        If no data is being collected, this is a no-op.
        """
        assert len(kwargs) == 1
        self.log(**kwargs)
        return next(iter(kwargs.values()))

    @contextlib.contextmanager
    def collect(self, enabled=True) -> Context:
        """Collect instrumentation data.

        This is a context manager that collects all instrumentation data
        that is recorded inside the context.
        """
        context = Context()

        if not enabled:
            yield context
            return

        outer = self._context
        self._thread_local.context = context
        try:
            yield context
            # If there is an outer, merge scope into it.
            if outer is not None:
                # Merge each key separately.
                for key, values in context.items():
                    outer.setdefault(key, []).extend(values)
        finally:
            self._thread_local.context = outer


class NullInstrumentation(Instrumentation):
    class _NullScope:
        def __enter__(self):
            pass

        def __exit__(self, *args):
            pass

        def __call__(self, func):
            return func

    _null_scope = _NullScope()

    @property
    def is_active(self) -> bool:
        return False

    def log(self, **kwargs):
        pass

    def spy(self, **kwargs):
        if len(kwargs) == 1:
            return next(iter(kwargs.values()))

    def scope(self, key: str):
        return NullInstrumentation._null_scope

    @contextlib.contextmanager
    def collect(self, enabled=True) -> Context:
        # Log a warning if enabled.
        if enabled:
            print("Warning: collecting data with NullInstrumentation")
        yield Context()


if os.getenv("NO_INSTRUMENTATION"):
    instrumentation = NullInstrumentation()
else:
    instrumentation = Instrumentation()



"""
Project metadata collector.

Collects different metadata (from git and wandb) to make it easy to pin down an experiment.
"""
from dataclasses import dataclass
from blackhc.project import project_dir


@dataclass
class ProjectMetadata:
    project_dir: str | None
    git_commit: str | None
    git_url: str | None
    wandb_project: str | None
    wandb_url: str | None


# https://chat.openai.com/share/5f3f2019-2051-4217-93ea-c926fa3c2749
try:
    import markdownify
except ImportError:
    raise ImportError("Package 'markdownify' is not installed. Please install it using 'pip install markdownify'")

try:
    from IPython.core.getipython import get_ipython
    from IPython.display import HTML
except ImportError:
    raise ImportError("Package 'IPython' is not installed. Please install it using 'pip install ipython'")

try:
    from rich import print
    from rich.console import Console
    from rich.markdown import Markdown
except ImportError:
    raise ImportError("Package 'rich' is not installed. Please install it using 'pip install rich'")


def print_html(obj):
    # Create a Rich console that outputs to a string
    console = Console()

    # Convert the HTML to Markdown
    markdown = markdownify.markdownify(obj.data)
    rich_markdown = Markdown(markdown)
    console.print(rich_markdown)


if __name__ == "__main__":
    # Create InteractiveShell instance
    from IPython.core.interactiveshell import InteractiveShell

    InteractiveShell.instance()


# Get the current IPython instance
ipython = get_ipython()

if ipython is not None:
    # Register the renderer for HTML objects
    ipython.display_formatter.ipython_display_formatter.for_type(HTML, print_html)


if __name__ == "__main__":
    print(
        ipython.display_formatter.format(
            HTML(
                "<h1>Hello World</h1><p>This is a paragraph</p><ul><li>Item 1</li><li>Item 2</li></ul>"
            )
        )
    )



from unittest.mock import patch

import pytest

import wandb
from blackhc.project.utils.log_path import xpath
from blackhc.project.utils.wandb_log_path import (
    commit,
    define_metric,
    log,
    log_metric,
    wandb_step,
)


@pytest.fixture(autouse=True)
def reset_xpath_state():
    # Reset the state of xpath to its initial state
    # This could be resetting variables or re-instantiating the object
    xpath.test_reset()

    yield  # This allows the test to run. After the test, any cleanup can occur if needed.


@pytest.mark.integration
def test_integration_wandb_log_path():
    run = wandb.init(project="dummy_project", mode="online")

    # Test wandb_step
    with wandb_step("test_step"):
        assert xpath.current_step_name == "test_step"
        assert xpath.current_step_index == 0

        # Test log_metric
        log_metric("test_metric", 1)

        # Test log
        log({"test_metric2": 2})

        # Test commit
        commit()
        # Check if commit was successful

        # Test define_metric
        define_metric("test_metric3")

    run.finish()


def test_wandb_log_path():
    with (
        patch("wandb.log"),
        patch("wandb.define_metric"),
        patch("wandb.finish"),
        patch("wandb.init"),
    ):
        wandb.init(project="dummy_project", mode="offline")

        assert xpath.path_separator == "/"

        # Test wandb_step
        with wandb_step("test_step"):
            wandb.define_metric.assert_called_with("test_step/*", "test_step/step")

            assert xpath.current_step_name == "test_step"
            assert xpath.current_step_index == 0
            wandb.log.assert_called_with({"test_step/step": 0}, commit=False)

            # Test log_metric
            log_metric("test_metric", 1)
            wandb.log.assert_called_with({"test_step/test_metric": 1}, commit=False)

            # Test log
            log({"test_metric2": 2})
            wandb.log.assert_called_with({"test_step/test_metric2": 2}, commit=False)

            # Test commit
            commit()
            # Check if commit was successful
            wandb.log.assert_called_with({}, commit=True)

            # Test define_metric
            define_metric("test_metric3")
            wandb.define_metric.assert_called_with("test_step/test_metric3")
            # Verify that except for the two calls there were no others
            assert wandb.define_metric.call_count == 2

        wandb.log.assert_called_with({}, commit=True)

        wandb.finish()



# Tests for the xpaths package

import pytest

from blackhc.project.utils.log_path import xpath


@pytest.fixture(autouse=True)
def reset_xpath_state():
    # Reset the state of xpath to its initial state
    # This could be resetting variables or re-instantiating the object
    xpath.test_reset()

    yield  # This allows the test to run. After the test, any cleanup can occur if needed.


def test_xpath_no_step():
    assert xpath.metric_name("metric") == "metric"

    # Check all metrics and paths
    assert xpath.all_metrics == {"metric"}
    assert xpath.all_step_names == set()


def test_xpath():
    with xpath.step("outer"):
        with xpath.step("inner"):
            assert xpath.metric_name("metric") == "outer/inner/metric"
        with xpath.step("inner2"):
            assert xpath.metric_name("metric") == "outer/inner2/metric"

    assert xpath.metric_name("metric") == "metric"

    # Check all metrics and paths
    assert xpath.all_metrics == {"metric", "outer/inner/metric", "outer/inner2/metric"}
    assert xpath.all_step_names == {"outer", "outer/inner", "outer/inner2"}


def test_xpath_array():
    with xpath.step("inner"):
        assert xpath.metric_name("metric", is_summary=True) == "inner[0]/metric"
    with xpath.step("inner"):
        assert xpath.metric_name("metric", is_summary=True) == "inner[1]/metric"


def test_xpath_missing_array():
    # Checks that a warning is logged if the same sub_name is used twice without as_array=True
    with pytest.warns(UserWarning):
        with xpath.step("inner", is_summary_step=True):
            pass
        with xpath.step("inner", is_summary_step=True):
            pass

    # Check all metrics and paths
    assert xpath.all_summary_step_names == {"inner", "inner+1"}


def test_xpath_current_step_name():
    with xpath.step("outer"):
        assert xpath.current_step_name == "outer"
        with xpath.step("inner"):
            assert xpath.current_step_name == "outer/inner"
        with xpath.step("inner"):
            assert xpath.current_step_name == "outer/inner"
        assert xpath.current_step_name == "outer"
    assert xpath.current_step_name == ""

    # Check all metrics and paths
    assert xpath.all_step_names == {"outer", "outer/inner"}


def test_xpath_current_unique_path():
    with xpath.step("outer"):
        assert xpath.current_unique_path == "outer[0]"
        with xpath.step("inner"):
            assert xpath.current_unique_path == "outer[0]/inner[0]"
        with xpath.step("inner"):
            assert xpath.current_unique_path == "outer[0]/inner[1]"
        assert xpath.current_unique_path == "outer[0]"
    assert xpath.current_step_name == ""


def test_xpath_all_paths_stats():
    with xpath.step("outer"):
        with xpath.step("inner"):
            with xpath.step("inner2"):
                with xpath.step("inner3"):
                    pass
    assert xpath.all_step_stats["outer/inner/inner2/inner3"].count == 1
    assert xpath.all_step_stats["outer/inner/inner2"].count == 1
    assert xpath.all_step_stats["outer/inner"].count == 1
    assert xpath.all_step_stats["outer"].count == 1

    assert (
        xpath.all_step_stats["outer/inner/inner2/inner3"].mean_time
        == xpath.all_step_stats["outer/inner/inner2/inner3"].total_time
    )
    assert (
        xpath.all_step_stats["outer/inner/inner2"].mean_time
        == xpath.all_step_stats["outer/inner/inner2"].total_time
    )
    assert (
        xpath.all_step_stats["outer/inner"].mean_time
        == xpath.all_step_stats["outer/inner"].total_time
    )
    assert (
        xpath.all_step_stats["outer"].mean_time
        == xpath.all_step_stats["outer"].total_time
    )

    with xpath.step("outer"):
        with xpath.step("inner"):
            with xpath.step("inner2"):
                with xpath.step("inner3"):
                    pass

    assert xpath.all_step_stats["outer/inner/inner2/inner3"].count == 2
    assert xpath.all_step_stats["outer/inner/inner2"].count == 2
    assert xpath.all_step_stats["outer/inner"].count == 2
    assert xpath.all_step_stats["outer"].count == 2

    assert xpath.all_step_stats["outer/inner/inner2/inner3"].mean_time > 0
    assert xpath.all_step_stats["outer/inner/inner2"].mean_time > 0
    assert xpath.all_step_stats["outer/inner"].mean_time > 0
    assert xpath.all_step_stats["outer"].mean_time > 0

    assert xpath.all_step_stats["outer/inner/inner2/inner3"].total_time > 0
    assert xpath.all_step_stats["outer/inner/inner2"].total_time > 0
    assert xpath.all_step_stats["outer/inner"].total_time > 0
    assert xpath.all_step_stats["outer"].total_time > 0


def test_xpath_summary_step():
    with xpath.step("outer", is_summary_step=True):
        with xpath.step("inner"):
            assert xpath.current_step_name == "outer/inner"
            assert xpath.current_unique_path == "outer/inner"
            assert xpath.metric_name("metric") == "outer/inner/metric"
        with pytest.warns(UserWarning):
            with xpath.step("inner"):
                assert xpath.current_step_name == "outer/inner+1"
                assert xpath.current_unique_path == "outer/inner+1"
                assert xpath.metric_name("metric") == "outer/inner+1/metric"
        assert xpath.current_step_name == "outer"
        assert xpath.current_unique_path == "outer"
        assert xpath.metric_name("metric") == "outer/metric"



from blackhc.project.utils.instrumentation import (
    Context,
    Instrumentation,
    NullInstrumentation,
)


def test_instrumentation_example():
    instrument = Instrumentation()

    with instrument.collect() as run_info:
        instrument.log(loss=1)

        with instrument.scope("my_scope"):
            instrument.log(info=2)
            assert instrument.spy(loss=3) == 3

    assert run_info == Context(loss=[1], my_scope=[Context(info=[2], loss=[3])])


def test_instrumentation_decorator():
    instrument = Instrumentation()

    @instrument.scope("my_scope")
    def my_func():
        instrument.log(info=2)

    with instrument.collect() as run_info:
        instrument.log(loss=1)
        my_func()

    assert run_info == Context(loss=[1], my_scope=[Context(info=[2])])


def test_instrumentation_nested_scopes():
    instrument = Instrumentation()

    with instrument.collect() as run_info:
        instrument.log(loss=1)

        with instrument.scope("my_scope"):
            instrument.log(info=2)

            with instrument.scope("my_scope"):
                instrument.log(info=3)

    assert run_info == Context(
        loss=[1],
        my_scope=[
            Context(
                info=[2],
                my_scope=[
                    Context(info=[3]),
                ],
            ),
        ],
    )


def test_instrumentation_nested_collects():
    instrument = Instrumentation()

    with instrument.collect() as run_info:
        instrument.log(loss=1)

        with instrument.collect() as run_info_inner:
            instrument.log(loss=2)

            with instrument.scope("my_scope"):
                instrument.log(info=3)

    assert run_info == Context(
        loss=[1, 2],
        my_scope=[
            Context(info=[3]),
        ],
    )

    assert run_info_inner == Context(
        loss=[2],
        my_scope=[
            Context(info=[3]),
        ],
    )


def test_instrumentation_walk():
    instrument = Instrumentation()

    with instrument.collect() as run_info:
        instrument.log(loss=1)

        with instrument.scope("my_scope"):
            instrument.log(info=2)

    assert list(run_info.walk()) == [
        ("loss", 1),
        ("my_scope", Context(info=[2])),
    ]


def test_instrumentation_tree_walk():
    instrument = Instrumentation()

    with instrument.collect() as run_info:
        instrument.log(loss=1)

        with instrument.scope("my_scope"):
            instrument.log(info=2)

    assert list(run_info.tree_walk()) == [
        (("loss",), 1),
        (("my_scope",), Context(info=[2])),
        (("my_scope", "info"), 2),
    ]

    # Check that we can pattern match as well
    for item in run_info.tree_walk():
        match item:
            case (["loss"], value):
                assert value == 1
            case (["my_scope"], context):
                assert context == Context(info=[2])
            case (["my_scope", "info"], value):
                assert value == 2


def test_instrumentation_to_plain_dict():
    instrument = Instrumentation()

    with instrument.collect() as run_info:
        instrument.log(loss=1)

        with instrument.scope("my_scope"):
            instrument.log(info=2)

    print(run_info)

    assert run_info.to_dict() == {
        "loss": [1],
        "my_scope": [
            {
                "info": [2],
            }
        ],
    }


def test_null_instrumentation():
    instrument = NullInstrumentation()

    with instrument.collect() as run_info:
        instrument.log(loss=1)

        with instrument.scope("my_scope"):
            instrument.log(info=2)

    assert run_info == Context()


if __name__ == "__main__":
    test_instrumentation_example()
    test_instrumentation_decorator()
    test_instrumentation_nested_scopes()
    test_instrumentation_nested_collects()
    test_instrumentation_walk()
    test_instrumentation_tree_walk()



"""
MIT License

Copyright (c) 2023 Andreas 'blackhc' Kirsch

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import operator
from dataclasses import dataclass


@dataclass
class EarlyStopping:
    """Early stopping generator.

    We can zip this generator with the training loop to stop training early.

    :param patience: how many epochs/upddates to wait for improvement.
    :param lower_is_better: whether a lower value is better.
    """

    patience: int
    lower_is_better: bool = True

    def __post_init__(self):
        self.max_patience = self.patience
        if self.lower_is_better:
            self.best_score = float("inf")
            self.cmp = operator.lt
        else:
            self.best_score = -float("inf")
            self.cmp = operator.gt

    @property
    def patience_remaining(self) -> bool:
        return self.patience >= 0

    def step(self, score):
        assert self.patience_remaining
        if self.cmp(score, self.best_score):
            self.best_score = score
            self.patience = self.max_patience
        else:
            self.patience -= 1

        return self.patience_remaining

    def __iter__(self):
        step_counter = 0
        while self.patience_remaining:
            yield self.step
            step_counter += 1
        print(
            f"Early stopping with best score {self.best_score} at step {step_counter}"
        )



"""
Simple file-based object storage.
"""
import functools
import inspect
import json
import os
import pickle
import sys
import typing
import urllib.parse

from datetime import datetime

from blackhc.project.experiment import get_git_head_commit_and_url


try:
    import wandb
except ImportError:
    wandb = None


try:
    import torch
except ImportError:
    torch = None


def get_module_name(f):
    """Get the name of the module of an object that has a __module__."""    
    module = sys.modules[f.__module__]
    return module.__spec__.name if hasattr(module, "__spec__") else module.__name__


def get_callable_full_name(f: typing.Callable):
    # return f"{get_module_name(f)}:{f.__qualname__}"
    return f.__qualname__


def escape_path_part(part: str):
    return urllib.parse.quote(part, safe=" ")


def arg_to_path_fragment(value: float | str | int | list[float | int | str]) -> str:
    """Convert a value to a path part."""
    if isinstance(value, float) and value.is_integer():
        value = int(value)
    if isinstance(value, float):
        value = format(
            value, ".6g"
        )  # Use general format with up to 6 significant digits
    elif isinstance(value, list):
        value = "[" + ",".join(map(arg_to_path_fragment, value)) + "]"
    elif isinstance(value, tuple):
        value = "(" + ",".join(map(arg_to_path_fragment, value)) + ")"
    elif isinstance(value, dict):
        value = (
            "{"
            + ",".join(
                f"{arg_to_path_fragment(k)}:{arg_to_path_fragment(v)}"
                for k, v in value.items()
            )
            + "}"
        )
    elif isinstance(value, int):
        value = str(value)
    elif isinstance(value, str):
        pass
    else:
        raise ValueError(f"Unsupported value type: {type(value)}")
    return escape_path_part(str(value))


def kwargs_to_fragment(kwargs: dict, incl_keys: bool = True):
    kwarg_fragments = []
    for key in sorted(kwargs.keys()):
        if incl_keys:
            kwarg_fragments.append(f"{key}:{arg_to_path_fragment(kwargs[key])}")
        else:
            kwarg_fragments.append(f"{arg_to_path_fragment(kwargs[key])}")
    return "_".join(kwarg_fragments)


def generate_path(*parts) -> str:
    """
    Generates a path based on the given identifier and keyword arguments.

    Args:
        identifier (str or callable): The identifier to be used in the path.
        **kwargs: Additional keyword arguments to be included in the path.

    Returns:
        str: The generated path.
    """
    path_parts = []
    for part in parts:
        if part is None:
            fragment = ''
        elif isinstance(part, dict):
            fragment = kwargs_to_fragment(part)
        else:
            fragment = arg_to_path_fragment(part)
        if fragment != '':
            path_parts.append(fragment)
            
    if parts and parts[-1] is None:
        path_parts.append('')
        
    return "/".join(path_parts)


def collect_metadata(*parts) -> dict[str]:
    head_commit, github_url = get_git_head_commit_and_url(os.getcwd())
    # If wandb is running, get the wandb id and url
    wandb_id = None
    wandb_url = None
    if wandb is not None and wandb.run is not None:
        wandb_id = wandb.run.id
        wandb_url = wandb.run.get_url()

    metadata = dict(
        timestamp=datetime.now().isoformat(),
        git=dict(commit=head_commit, url=github_url),
        wandb=dict(id=wandb_id, url=wandb_url),
        parts=parts,
    )
    return metadata


def get_prefix_path(*parts, root: str = "") -> str:
    return os.path.join(root, generate_path(*parts))


def _combine_path(prefix_path, ext) -> str:
    if prefix_path.endswith("/"):
        return f"{prefix_path}{ext}"
    return f"{prefix_path}.{ext}"


def _save_metadata(*parts, root: str = "") -> str:
    prefix_path = get_prefix_path(*parts, root=root)
    metadata = collect_metadata(*parts)
    os.makedirs(os.path.dirname(prefix_path), exist_ok=True)
    with open(_combine_path(prefix_path, "meta.json"), "wt", encoding="utf-8") as f:
        json.dump(metadata, f)
    return prefix_path


def load_metadata(*parts, root: str = ""):
    prefix_path = get_prefix_path(*parts, root=root)
    with open(_combine_path(prefix_path, "meta.json"), "rt", encoding="utf-8") as f:
        return json.load(f)


def save_pkl(obj, *parts, root: str = "") -> str:
    prefix_path = _save_metadata(*parts, root=root)
    with open(_combine_path(prefix_path, "data.pkl"), "wb") as f:
        pickle.dump(obj, f)
    return prefix_path


def load_pkl(*parts, root: str = ""):
    prefix_path = get_prefix_path(*parts, root=root)
    with open(_combine_path(prefix_path, "data.pkl"), "rb") as f:
        return pickle.load(f)


def save_pt(obj, *parts, root: str = "") -> str:
    prefix_path = _save_metadata(*parts, root=root)
    with open(_combine_path(prefix_path, "data.pt"), "wb") as f:
        torch.save(obj, f)
    return prefix_path


def load_pt(*parts, root: str = ""):
    prefix_path = get_prefix_path(*parts, root=root)
    with open(_combine_path(prefix_path, "data.pt"), "rb") as f:
        return torch.load(f)


def save_json(obj, *parts, root: str = "") -> str:
    prefix_path = _save_metadata(*parts, root=root)
    with open(_combine_path(prefix_path, "data.json"), "wt", encoding="utf-8") as f:
        json.dump(obj, f)
    return prefix_path


def load_json(*parts, root: str = ""):
    prefix_path = get_prefix_path(*parts, root=root)
    with open(_combine_path(prefix_path, "data.json"), "rt", encoding="utf-8") as f:
        return json.load(f)
    
    
def save_pkl_or_json(obj, *parts, root: str = "") -> str:
    prefix_path = _save_metadata(*parts, root=root)
    
    # Pickle the object into bytes
    pickled_obj = pickle.dumps(obj)
    
    # Check if the size is less than 256KB
    if len(pickled_obj) < 256 * 1024:
        try:
            # Try to save as JSON
            json_obj = json.loads(json.dumps(obj))
            assert json_obj == obj
            with open(_combine_path(prefix_path, "data.json"), "wt", encoding="utf-8") as f:
                json.dump(obj, f)
            return prefix_path
        except (TypeError, OverflowError, AssertionError):
            # If it fails, save as pickle
            pass
   
    with open(_combine_path(prefix_path, "data.pkl"), "wb") as f:
        f.write(pickled_obj)
    return prefix_path


def load(*parts, root: str = ""):
    prefix_path = get_prefix_path(*parts, root=root)
    
    # Find the *data.* file (can either end in pkl, json or ot)
    data_files = [f for f in os.listdir(os.path.dirname(prefix_path)) if f.endswith(("data.pt", "data.pkl", "data.json"))]
    
    if len(data_files) == 1:
        data_file = os.path.join(os.path.dirname(prefix_path), data_files[0])
        if data_file.endswith(".pkl"):
            with open(data_file, "rb") as f:
                return pickle.load(f)
        elif data_file.endswith(".json"):
            with open(data_file, "rt", encoding="utf-8") as f:
                return json.load(f)
        elif data_file.endswith(".pt"):
            with open(data_file, "rb") as f:
                return torch.load(f)
        else:
            raise ValueError(f"Unsupported file type: {data_file}")
    elif len(data_files) > 1:
        raise RuntimeError("Multiple data files found for the same prefix path", data_files)
    else:
        raise FileNotFoundError("No data file found for the prefix path", prefix_path)


def cache(f=None, *, prefix_args: list[str], root:str, force_format:typing.Literal["json", "pkl", "pt"] | None = None):
    if f is None:
        return functools.partial(cache, prefix_args=prefix_args, root=root, force_format=force_format)
    
    @functools.wraps(f)
    def f_get_prefix_path(*args, **kwargs):
         # Apply defaults
        sig = inspect.signature(f)
        bound_args = sig.bind(*args, **kwargs)
        bound_args.apply_defaults()
        
        # Extract prefix args into one dictionary and all other args into another
        prefix_dict = {arg: bound_args.arguments[arg] for arg in bound_args.arguments if arg in prefix_args}
        suffix_dict = {arg: bound_args.arguments[arg] for arg in bound_args.arguments if arg not in prefix_args}
        
        # Generate the prefix path
        prefix_path = get_prefix_path(prefix_dict, get_callable_full_name(f), suffix_dict, root=root)
        return prefix_path

    @functools.wraps(f)
    def f_load(*args, **kwargs):
        # Apply defaults
        sig = inspect.signature(f)
        bound_args = sig.bind(*args, **kwargs)
        bound_args.apply_defaults()
        
        # Extract prefix args into one dictionary and all other args into another
        prefix_dict = {arg: bound_args.arguments[arg] for arg in bound_args.arguments if arg in prefix_args}
        suffix_dict = {arg: bound_args.arguments[arg] for arg in bound_args.arguments if arg not in prefix_args}
        
        # Generate the prefix path
        prefix_path = get_prefix_path(prefix_dict, get_callable_full_name(f), suffix_dict, root=root)
        
        # Find all subdirs in the prefix path
        subdirs = sorted([d for d in os.listdir(prefix_path) if os.path.isdir(os.path.join(prefix_path, d))])
        if not subdirs:
            raise FileNotFoundError("No subdirectories found in the prefix path", prefix_path)
        latest_subdir = subdirs[-1]
        latest_subdir_path = os.path.join(prefix_path, latest_subdir)
        return load(root=latest_subdir_path)
    
    @functools.wraps(f)
    def f_recompute(*args, **kwargs):
        # Apply defaults
        sig = inspect.signature(f)
        bound_args = sig.bind(*args, **kwargs)
        bound_args.apply_defaults()
        
        # Extract prefix args into one dictionary and all other args into another
        prefix_dict = {arg: bound_args.arguments[arg] for arg in bound_args.arguments if arg in prefix_args}
        suffix_dict = {arg: bound_args.arguments[arg] for arg in bound_args.arguments if arg not in prefix_args}
            
        # No result has been cached. So execute the function.
        timestamp = datetime.now().isoformat()
        result = f(*bound_args.args, **bound_args.kwargs)
        
        match force_format:
            case "json":
                save_fn = save_json
            case "pkl":
                save_fn = save_pkl
            case "pt":
                save_fn = save_pt
            case None:
                if torch is not None and isinstance(result, torch.Tensor):
                    save_fn = save_pt
                else:
                    save_fn = save_pkl_or_json
            case _:
                raise ValueError(f"Unsupported force_format: {force_format}")
            
        save_fn(result, prefix_dict, get_callable_full_name(f), suffix_dict, timestamp, None, root=root)
        
        return result
    
    @functools.wraps(f)
    def f_wrapper(*args, **kwargs):
        try:
            result = f_load(*args, **kwargs)
            return result
        except FileNotFoundError:
            pass
        
        return f_recompute(*args, **kwargs)
        
    f_wrapper.get_prefix_path = f_get_prefix_path
    f_wrapper.load = f_load
    f_wrapper.recompute = f_recompute
    return f_wrapper


"""
This module provides a TreeNamespace class and related utility functions for working with
hierarchical data structures.

The TreeNamespace class allows for easy manipulation and access of nested dictionary-like
structures using dot notation and path-based indexing. It supports wildcards for flexible
querying of the data structure.

Key components:
- TreeNamespace: Main class for representing and interacting with tree-like data structures.
- get_generic_type: Function to determine the generic type of an object.
- keys, values, items: Functions to access elements of TreeNamespace and other iterable objects.
- schema: Function to generate a schema representation of a TreeNamespace or other objects.

This module also includes pretty-printing support for TreeNamespace objects using the
prettyprinter library.

Usage example:
```
    from blackhc.project.utils.tree_namespace import TreeNamespace, schema

    data = {
        "user/profile/name": "Alice",
        "user/profile/age": 30,
        "user/settings/theme": "dark",
        "posts/0/title": "First Post",
        "posts/1/title": "Second Post",
    }

    tree = TreeNamespace(data)

    # Access nested data using dot notation
    print(tree.user.profile.name)  # Output: Alice

    # Use wildcards to access multiple items
    post_titles = tree["posts/*/title"]
    print(post_titles["0"])  # Output: First Post

    # Get schema of the tree
    tree_schema = schema(tree)
    print(tree_schema.user.profile.age)  # Output: <class 'int'>
```

Note: This module requires the 'prettyprinter' package to be installed.
"""
import re
import typing
from typing import Any

try:
    import prettyprinter
except ImportError:
    raise ImportError("Package 'prettyprinter' is not installed. Please install it using 'pip install prettyprinter'")


class TreeNamespace:
    """A namespace for tree dicts.

    It stores a flat dictionary of key paths that map to values.
    Key paths are keys separated by / as strings.

    Every path can be accessed as an attribute, e.g. namespace.a.b.c for the key "a/b/c".
    Paths can also be accessed with the __getitem__ method, e.g. namespace["a/b/c"].

    A path that does not point to a value returns a new namespace with all keys that start
    with the path. For example, namespace["a"] returns a new namespace with all keys that
    start with "a/".

    The __getitem__ method also supports wildcards: namespace["a/*"] returns a new namespace
    with all keys that start with "a/". The wildcard * can be used to match any single key.
    The wildcard @ can be used to match any sub-path. For example, namespace["a/@/b"] returns
    a new namespace with all keys that start with "a/" and end with "/b".
    """

    def __init__(self, wrapped_dict):
        self._dict = wrapped_dict

    def __getattribute__(self, name: str) -> typing.Any:
        if name.startswith("_"):
            return super().__getattribute__(name)
        return self[name]

    def _schema(self) -> "TreeNamespace":
        """Get the schema of the namespace, which has the same keys
        but maps to the generic types."""
        schema_dict = {}
        for key, value in self._dict.items():
            if isinstance(value, TreeNamespace):
                sub_schema = value._schema()._dict
                for sub_key, sub_value in sub_schema.items():
                    schema_dict[key + "/" + sub_key] = sub_value
            else:
                schema_dict[key] = get_generic_type(value)
        return TreeNamespace(schema_dict)

    def _keys(self) -> list[str]:
        """Get the direct keys of the namespace as list."""
        # Note: sets are not ordered, but dicts are ordered in Python 3.7+
        keys = {key.split("/")[0]: None for key in self._dict.keys()}
        return list(keys.keys())

    def _values(self) -> list[Any]:
        """Get the direct values of the namespace as list.

        Sub-namespaces are of type TreeNamespace, too.
        """
        return [self[key] for key in self._keys()]

    def _items(self) -> list[tuple[str, Any]]:
        """Get the direct items of the namespace as list.

        Sub-namespaces are of type TreeNamespace, too.
        """
        return [(key, self[key]) for key in self._keys()]

    def __dir__(self) -> list[str]:
        return [*super().__dir__(), *self._keys()]

    def __getitem__(self, name: str) -> typing.Any:
        """Get a value from the namespace.

        * can be used as a wildcard to match any single key.
        @ can be used as a wildcard to match any sub-path.

        The returned value is either a single value or a new namespace if wild-cards are used.
        A single value can be a namespace itself if the key points to a sub-namespace.
        """
        name = str(name)
        wrapped_dict = self._dict
        if name in wrapped_dict:
            return wrapped_dict[name]

        # Check if there are any * in the name. If so, return a list of all
        # matching keys in a new namespace that only contains the matches within the
        # the key names.
        if "*" in name or "@" in name:
            # turn name into a regex: replace * with (.*) and it has to end with an optional /
            name_pattern = (
                "^"
                + name.replace("*", "([^/]*)").replace("@", "(.*?)")
                + "(?P<remainder>/.*|$)"
            )
            name_re = re.compile(name_pattern)

            # find all keys and their matches
            key_map = {}
            for key in wrapped_dict.keys():
                match = name_re.match(key)
                if not match:
                    continue

                new_parts = [group.replace("/", "_") for group in match.groups()[:-1]]
                remainder = match.groupdict()["remainder"]
                key_map[key] = "/".join(new_parts) + remainder
        else:
            # Find all keys that start with "name/" and return them as a namespace.
            # This is useful for accessing nested keys.
            key_map = {
                key: key[len(name) + 1 :]
                for key in wrapped_dict.keys()
                if key.startswith(name + "/")
            }
        if len(key_map) == 0:
            raise KeyError(f"Key '{name}' not found in {set(wrapped_dict.keys())}")
        return TreeNamespace(
            {mapped_key: wrapped_dict[key] for key, mapped_key in key_map.items()}
        )

    def __simple__repr__(self):
        if len(self._keys()) == 0:
            return {}
        elif (keys := self._keys())[0] == "0" and keys == list(
            map(str, range(len(keys)))
        ):
            return [
                value.__simple__repr__() if isinstance(value, TreeNamespace) else value
                for value in self._values()
            ]
        else:
            return {
                key: self[key].__simple__repr__()
                if isinstance(self[key], TreeNamespace)
                else self[key]
                for key in self._keys()
            }

    def __repr__(self):
        return f"TreeNamespace({self.__simple__repr__()})"

    def __len__(self):
        """Get the number of keys without / or the number of unique prefixes."""
        return len(self._keys())

    def __eq__(self, other):
        """Check if two TreeNamespace objects are equal."""
        if isinstance(other, self.__class__):
            return self._dict == other._dict
        elif isinstance(other, dict):
            return self._dict == other or all(
                left == right for left, right in zip(self._items(), other.items())
            )
        return False

    def __hash__(self):
        """Return the hash of the TreeNamespace object."""
        return hash(set(self.__dict__.items()))


def get_generic_type(obj: Any) -> type:
    """Get the generic type of an object.

    For lists, tuples, and dicts, this returns the type of the elements.
    """
    if isinstance(obj, (list, tuple)):
        if len(obj) == 0:
            return type(obj)
        test_type = get_generic_type(obj[0])
        if any(get_generic_type(elem) != test_type for elem in obj):
            return type(obj)
        return type(obj)[test_type]
    if isinstance(obj, dict):
        if len(obj) == 0:
            return type(obj)
        test_key_type = get_generic_type(next(iter(obj.keys())))
        test_value_type = get_generic_type(next(iter(obj.values())))
        if any(get_generic_type(key) != test_key_type for key in obj.keys()) or any(
            get_generic_type(value) != test_value_type for value in obj.values()
        ):
            return type(obj)
        return type(obj)[test_key_type, test_value_type]
    return type(obj)


@prettyprinter.register_pretty(TreeNamespace)
def pretty_deep_key_namespace(value, ctx):
    return prettyprinter.pretty_call(
        ctx,
        TreeNamespace,
        **value.__simple__repr__(),
    )


def keys(obj: Any) -> list[str]:
    """Get all keys of an object.

    For lists, tuples, and dicts, this returns the keys of the elements.
    """
    if isinstance(obj, TreeNamespace):
        return obj._keys()
    elif isinstance(obj, dict):
        return list(obj.keys())
    elif isinstance(obj, (list, tuple)):
        return list(map(str, range(len(obj))))
    else:
        raise TypeError(f"Cannot get keys of type {type(obj)}")


def values(obj: Any) -> list[Any]:
    """Get all values of an object.

    For lists, tuples, and dicts, this returns the values of the elements.
    """
    if isinstance(obj, TreeNamespace):
        return obj._values()
    elif isinstance(obj, dict):
        return list(obj.values())
    elif isinstance(obj, (list, tuple)):
        return list(obj)
    else:
        raise TypeError(f"Cannot get values of type {type(obj)}")


def items(obj: Any) -> list[tuple[str, Any]]:
    """Get all items of an object.

    For lists, tuples, and dicts, this returns the items of the elements.
    """
    if isinstance(obj, TreeNamespace):
        return obj._items()
    elif isinstance(obj, dict):
        return list(obj.items())
    elif isinstance(obj, (list, tuple)):
        return list(enumerate(obj))
    else:
        raise TypeError(f"Cannot get items of type {type(obj)}")


def schema(obj: Any) -> TreeNamespace | type:
    """Get the schema of an object.

    For lists, tuples, and dicts, this returns the schema of the elements.
    """
    if isinstance(obj, TreeNamespace):
        return obj._schema()
    elif isinstance(obj, dict):
        return TreeNamespace({key: schema(value) for key, value in obj.items()})
    elif isinstance(obj, (list, tuple)):
        return TreeNamespace({str(key): schema(value) for key, value in enumerate(obj)})
    else:
        return get_generic_type(obj)



# Always prefer setuptools over distutils
from setuptools import setup

# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="blackhc.project",
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version="4.1.1",
    description="Notebook setup code",
    # Fix windows newlines.
    long_description=long_description.replace("\r\n", "\n"),
    long_description_content_type="text/markdown",
    # The project's main homepage.
    url="https://github.com/blackhc/notebook_setup",
    # Author details
    author="Andreas @blackhc Kirsch",
    author_email="blackhc+notebook_setup@gmail.com",
    # Choose your license
    license="MIT",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 5 - Production/Stable",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Software Development :: Libraries :: Python Modules",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.10",
    ],
    # What does your project relate to?
    keywords="jupyter",
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=[
        "blackhc.project",
        "blackhc.project.utils",
        "blackhc.project.utils.tests",
    ],
    package_dir={"": "src"},
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        "gitpython",
        "fs.sshfs",
        "laaos",
        "toolz",  # Functional programming
        # See https://x.com/BlackHC/status/1780995421852127477.
        "jupyter_client<8.0.0",  # This fixes issues with jupyter_client <-> pyzmq
        "pyzmq<25",  # This fixes issues with jupyter_client <-> pyzmq
    ],
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e ".[dev,test]"
    extras_require={
        "dev": ["check-manifest"],
        "test": [
            "coverage",
            "pytest",
            "pytest-forked",
            "pyfakefs==3.3",
            "torch",
            "psutil",
            "pytorch-ignite",
            "blackhc.progress_bar",
            "wandb",
            "prettyprinter",
            "rich",
            "markdownify",
            "ipython",
        ],
    },
    setup_requires=[
        "pytest-runner",
    ],
)
